# Message Passing Interface

In order to realize distributed programming, the standard (+MPI) was created in 1993-1994 to standardize the passage of messages between several computers or in a computer with several processors/cores [@noauthor_message_2021]. (+MPI) is, therefore, a communication protocol and not a programming language. Currently, the latest version of (+MPI) is 4.0 which approved in 2021. There are several implementations of the standard:

* MPICH, which support for the moment, MPI 3.1,
* Open MPI, which support, for the moment, MPI 3.1

We use Open MPI throughout this project to distribute a cellular automaton across multiple compute nodes and exchange missing neighbors of a cell.

\pagebreak

## Example by imitating a token ring network

To understand the basis of (+MPI), let us look at an example mimicking a *token ring* network [@kendall_mpi_2018]. This type of network forces a process to send a message to the message in the console, for example, only if it has the token in its possession. Moreover, once it has emitted its message, the process must transmit the token to its neighbor.

\cimgl{figs/ring.png}{scale=0.4}{Imitation of a network in \textit{token ring}}{Source: Created by Baptiste Coudray}{fig:ex-mpi}

In this example (\ref{fig:ex-mpi}), the node with the rank zero has first the token that it will pass to node one, then it will give it to node two, and so on. The program ends when the token is back in possession of the process zero: node four sends the token to node zero.

```c
int main(int argc, char** argv) {
    // Initialize the MPI environment
    MPI_Init(&argc, &argv);
    // Find out rank, size
    int world_rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
    int world_size;
    MPI_Comm_size(MPI_COMM_WORLD, &world_size);
    
    int token;
    // Receive from the lower process and send to the higher process. Take care
    // of the special case when you are the first process to prevent deadlock.
    if (world_rank != 0) {
        MPI_Recv(&token, 1, MPI_INT, world_rank - 1, 0, MPI_COMM_WORLD, 
                 MPI_STATUS_IGNORE);
        printf("Process %d received token %d from process %d\n", 
               world_rank, token, world_rank - 1);
    } else {
        // Set the token's value if you are process 0
        token = -1;
    }
    MPI_Send(&token, 1, MPI_INT, (world_rank + 1) % world_size, 0,
    MPI_COMM_WORLD);
    // Now process 0 can receive from the last process. This makes sure that at
    // least one MPI_Send is initialized before all MPI_Recvs (again, to prevent
    // deadlock)
    if (world_rank == 0) {
        MPI_Recv(&token, 1, MPI_INT, world_size - 1, 0, MPI_COMM_WORLD,
        MPI_STATUS_IGNORE);
        printf("Process %d received token %d from process %d\n", world_rank, 
               token, world_size - 1);
    }
    return MPI_Finalize();
}
```

Any parallel program using (+MPI) must call the `MPI_Init` function to initialize the environment, otherwise, an error message is displayed when using another MPI function. Then, `MPI_Comm_rank` allows us to retrieve our ID (the node number we have), and then with the `MPI_Comm_size` function, we get the number of nodes on which our program is running, in this case five.
Thanks to the node number, the node with the identifier zero, sends the token to its neighbor via the `MPI_Send` function.

So, once sent, it waits for node four to send the token via the function `MPI_Recv`. Then, the other nodes are waiting to receive the token from their neighbor to pass the token in turn.
The nodes communicate through the communicator `MPI_COMM_WORLD`, a macro-constant designating all nodes associated with the current program.

The function `MPI_Send` and `MPI_Recv` can be blocking or non-blocking, depending on the size of the message that we send. Indeed, there is an internal message buffer; while it is not full, the functions will be non-blocking; otherwise, they will be blocking.

Finally, every program must terminate with the `MPI_Finalize` function; otherwise, the execution ends with an error message.

```bash
mpicc ring.c -o ring
mpirun -n 5 ./ring
```

To compile a (+MPI) program, you have to go through the `mpicc` program, which is a wrapper
around (+GCC). Indeed, `mpicc` automatically adds the correct compilation parameters to the (+GCC) program.
Next, our compiled program must be run through `mpirun` to distribute our program to compute nodes. Finally, the `-n` parameter is used to specify the number of processes to run.

```
Process 1 received token -1 from process 0
Process 2 received token -1 from process 1
Process 3 received token -1 from process 2
Process 4 received token -1 from process 3
Process 0 received token -1 from process 4
```
Thus, we can see that the processes exchange the token each in turn until node zero receives the token again.

\pagebreak
