# Cellular Automaton

A cellular automaton consists of a regular grid of cells, each in one of a finite number of states. The grid can be in any finite number of dimensions. For each cell, a set of cells called its neighborhood is defined relative to the specified cell. An initial state (time $t = 0$) is selected by assigning a state for each cell. A new generation is created (advancing t by 1), according to some fixed rule (generally, a mathematical function) that determines the new state of each cell in terms of the current state of the cell and the states of the cells in its neighborhood. Typically, the rule for updating the state of cells is the same for each cell and does not change over time [@noauthor_automate_2021].

The neighborhood of a cell is defined either by the Moore neighborhood or by the Von Neumann neighborhood. The first one defines that a cell has in a two-dimensional cellular automaton four neighbors while the second one, eight.

\cimg{figs/neighbours.png}{scale=0.5}{Comparison between Von Neumann (left) and Moore (right) neighborhoods}{Source: Created by Baptiste Coudray}

The grid on the left represents the Von Neumann neighborhood, i.e., the four neighbors of a cell. These are denoted by the four cardinal points (north, west, south, east).
The grid on the right represents Moore's neighborhood, i.e., the eight neighbors of a cell. These are denoted by the four cardinal points and the four inter-cardinal points (northwest, southwest, southeast, northeast).

The cellular automaton will have to use the Moore neighborhood, which means that each cell has:

* two neighbors in one dimension,
* eight neighbors in two dimensions,
* 26 neighbors in three dimensions.

These values are valid for a cellular automaton of dimension two and a Chebyshev distance of one. We can generalize the number of neighbors that a cell has via the formula $(2r + 1)^d - 1$, where $r$ is the Chebyshev distance, and $d$ is the dimension.

## MPI x Futhark

Our library allows distributing cellular automata automatically so that the programmer only has to write the Futhark function to update his cellular automaton. Our library supports cellular automata of one, two, and three dimensions and with any types of data. The use of the Futhark language allows to quickly update the state of the cellular automaton thanks to the different backend available. Therefore, several modes are available:

* distributed-sequential, the Futhark code executes sequentially,
* distributed-multicore, the Futhark code executes concurrently to (+POSIX) threads,
* distributed-OpenCL/(+CUDA), the Futhark code executes on the graphics card.

### Communication

Communication between the different MPI tasks is necessary to recover the missing neighbors and recreate the complete cellular automaton. Therefore, we create a virtual Cartesian topology. 
"*A virtual topology is a mechanism for naming the processes in a communicator in away that fits the communication pattern better. The main aim of this is to make sub-sequent code simpler. It may also provide hints to the run-time system which allow it to optimise the communication or even hint to the loader how to configure the processes. The virtual topology might also gain us some performance benefit.*" [@macdonald_writing_nodate]

\pagebreak

#### One dimension

\cimgl{figs/communication_1d.png}{scale=0.60}{Example of Cartesian virtual topology in one dimension}{Source: Created by Baptiste Coudray}{fig:comm-1d}

In a one-dimensional Cartesian topology like \ref{fig:comm-1d}, we notice that the rows can communicate directly with their left and right neighbors even if they are at the ends of the network. Indeed, the MPI communicator is defined as cyclic, which avoids having to traverse the $N - 2$ neighbors that separate them. 

#### Two dimensions

\cimgl{figs/communication_2d.png}{scale=0.60}{Example of Cartesian virtual topology in two dimensions}{Source: Created by Baptiste Coudray}{fig:comm-2d}

In a two-dimensional Cartesian topology like \ref{fig:comm-2d}, we notice that each rank can communicate directly with their left, right, top, and bottom neighbors. For each row there is an MPI communicator containing the ranks available in it:

* first communicator has `Rank 0` and `Rank 1`,
* second communicator has `Rank 2` and `Rank 3`.

For each column, there is an MPI communicator containing the ranks available in it:

* first communicator has `Rank 0` and `Rank 2`,
* second communicator has `Rank 1` and `Rank 3`.

Each communicator created is cyclic, so the first rank in each the communicator created can communicate with the last rank available in it. 

When a rank needs to communicate with its diagonal neighbor, we use the default communicator (`MPI_COMM_WORLD`) to communicate directly with each other without going through a neighbor.

### Three dimensions

\cimgl{figs/communication_3d.png}{scale=0.60}{Example of Cartesian virtual topology in three dimensions}{Source: Created by Baptiste Coudray}{fig:comm-3d}

In a three-dimensional Cartesian topology like \ref{fig:comm-3d}, we notice that the rows have the same communication capabilities as a two-dimensional topology, but, in addition, they can communicate with their front and back neighbors. Thus, two depth communicators are created:

* one containing `Rank 0` and `Rank 1`,
* one containing `Rank 2` and `Rank 3`.

### Data dispatching

The cellular automaton is shared as equally possible among the available tasks to perform more or less the same amount of work. Thus, each task has a chunk that is a part of the cellular automaton. This chunk match the dimension of the cellular automaton, so it can be of dimensions one, two, or three.

\pagebreak

#### One dimension

\cimgl{figs/dispatch_1d.png}{scale=0.50}{Example of sharing a cellular automaton in one dimension}{Source: Created by Baptiste Coudray}{fig:dd-1d}

In this example (\ref{fig:dd-1d}), a cellular automaton of dimension one, size eight, is split between three processes. As the division of the cellular automaton is not an integer, rank two have only two cells, unlike the others, which have three.

#### Two dimensions
\cimgl{figs/dispatch_2d.png}{scale=0.50}{Example of sharing a cellular automaton in two dimensions}{Source: Created by Baptiste Coudray}{fig:dd-2d}

In this example (\ref{fig:dd-2d}), the cellular automaton is in two dimensions and of size $6 \times 6$. With four tasks available, it can be separated into four sub-matrices of $3 \times 3$.

\pagebreak

#### Three dimensions

\cimgl{figs/dispatch_3d.png}{scale=0.40}{Example of sharing a cellular automaton in three dimension}{Source: Created by Baptiste Coudray}{fig:dd-3d}

In this example (\ref{fig:dd-3d}), the cellular automaton is in three dimensions and of size $4 \times{} 4 \times{} 2$. With four tasks available, it can be separated into four sub-cubes of $2 \times{} 2 \times{} 2$.

### Envelope

The envelope of a chunk represents the missing neighbours (at a Chebyshev distance chose by the programmer) of the cells at the extremities of the chunk. These missing cells are needed to compute the next iteration of the chunk of the cellular automaton that the process has.

#### One dimension

\cimgl{figs/envelope_1d.png}{scale=0.50}{Example of the envelope of a chunk in one dimension}{Source: Created by Baptiste Coudray}{fig:env-1d}

Using this one dimension cellular automaton (\ref{fig:dd-1d}), the Moore neighborhood of a cell includes the west-neighbor and the east-neighbor. We notice that the envelope of $R_{n}$ includes the last cell of $R_{(n-1)\:\%\:N}$ and the first cell of $R_{(n+1)\:\%\:N}$. 

For example, the envelope for R0 is the west-neighbor of one, so eight and the east neighbor of three, so four. The envelope for R1 is the west-neighbor of four, so three and the east neighbor of six, so seven. Finally, the envelope for R2 is the west-neighbor of seven, so six and the east neighbor of eight, so one. As we can see, the Chebyshev distance of all these envelopes is one. The ranks exchange data via MPI using the Cartesian virtual topology.

#### Two dimensions

\cimgl{figs/envelope_2d.png}{scale=0.50}{Example of the envelope of a chunk in two dimensions}{Source: Created by Baptiste Coudray}{fig:env-2d}

Using the two-dimensional cellular automaton described above (\ref{fig:dd-2d}), the chunk envelope of R0 requires eight communications. This example uses a Cartesian topology of size $2\times 2$ (m\times n), the neighbors are recovered as follows:

1. `North West Neighbors` are sent by $R_{ ((y - 1)\:\%\:m,\:(x - 1)\:\%\:n) }$,
2. `North Neighbors`, are sent by $R_{ ((y - 1)\:\%\:m,\:x) }$,
3. `North East Neighbors`, are sent by $R_{ ((y - 1)\:\%\:m,\:(x + 1)\:\%\:n) }$,
4. `East Neighbors`, are sent by $R_{ (y,\:(x + 1)\:\%\:n) }$,
5. `South East Neighbors`, are sent by $R_{ ((y + 1)\:\%\:m,\:(x + 1)\:\%\:n) }$,
6. `South Neighbors`, are sent by $R_{ ((y + 1)\:\%\:m,\:x) }$,
7. `South West Neighbors`, are sent by $R_{ ((y + 1)\:\%\:m,\:(x - 1)\:\%\:n) }$,
8. `West Neighbors`, are sent by $R_{ (y,\:(x - 1)\:\%\:n) }$.

As we can see, the Chebyshev distance of this envelope is one.

#### Three dimensions

With a three-dimensional cellular automaton like (\ref{fig:dd-3d}, the envelope of a chunk requires 26 MPI communications because we send:

* the eight vertices,
* the twelve edges,
* and the six faces around the chunk. 

\pagebreak
