# Simple Cellular Automaton

The simplest non-trivial cellular automaton that can be conceived consists of a one-dimensional grid of cells that can take only two states ("0" or "1"), with a neighborhood consisting, for each cell, of itself and the two cells adjacent to it [@noauthor_automate_2021].

There are $2^3 = 8$ possible configurations (or patterns, rules) of such a neighborhood. In order for the cellular automaton to work, it is necessary to define what the state must be at the next generation of a cell for each of these patterns. The eight rules/configurations defined is as follows:

| Rule n° | East neighbour state | Cell state | West neighbour state | Cell next state |
|:---:|:---:|:---:|:---:|:---:|
| 1 | 0 | 0 | 0 | 0
| 2 | 0 | 0 | 1 | 1
| 3 | 0 | 1 | 0 | 1
| 4 | 0 | 1 | 1 | 1
| 5 | 1 | 0 | 0 | 1
| 6 | 1 | 0 | 1 | 0
| 7 | 1 | 1 | 0 | 0
| 8 | 1 | 1 | 1 | 0
Table: Evolution rules for a cellule in a one dimensional cellular-automaton

\pagebreak

## Example

\cimg{figs/simple_automate.png}{scale=0.5}{First and second state of a SCA}{Source: Created by Baptiste Coudray}

Iteration 0 is the initial state and only cell two is alive. To perform the next iteration:

* the cell (one) is born because of rule n°2,
* the cell (two) stays alive because of rule n°3,
* the cell (three) stays alive because of rule n°6.

## Distributed version

With the created library, we implement this (+SCA) previously described. To do this, we create a Futhark `elementary.fut` file, which is used to calculate the next state of a part of the cellular automaton.

```
let get_neighbours [n] (elem: [n]i8) : [n](i8,i8) = ...

entry next_chunk_elems [n] (chunk_elems :[n]i8) :[]i8 =
    let neighbours = get_neighbours chunk_elems
    let next_elems = ...
    in next_elems[1:n-1]
```

Therefore, the `elementary.fut` file contains only a function that applies the rules on the cellular automaton.
As we can see, `next_chunk_elems` is the primary function that takes a chunk of the cellular automaton as a parameter.
This function applies the rules defined before on every cell and returns the new value of each cell without the envelope.

```c
void init_chunk_elems(chunk_info_t *ci) {
    int8_t *data8 = ci->data;
    for (int i = 0; i < ci->dimensions[1]; ++i) {
        data8[i] = rand() % 2;
    }
}
```

```c
void compute_next_chunk_board(struct dispatch_context *dc, 
        struct futhark_context *fc, chunk_info_t *ci) {
    struct futhark_i8_1d *fut_chunk_with_envelope = 
            get_chunk_with_envelope(dc, fc, 1, futhark_new_i8_1d);
    
    struct futhark_i8_1d *fut_next_chunk_elems;
    futhark_entry_next_chunk_elems(fc, &fut_next_chunk_elems, 
                                   fut_chunk_with_envelope);
    futhark_context_sync(fc);
    
    futhark_values_i8_1d(fc, fut_next_chunk_elems, ci->data);
    futhark_context_sync(fc);

    /* ... Free resources ... */
}

int main(int argc, char *argv[]) {
    /* ... MPI & Futhark Init ... */
    const int N_ITERATIONS = 100;
    int elems_dimensions[1] = {600};
    struct dispatch_context *disp_context = 
            dispatch_context_new(elems_dimensions, MPI_INT8_T, 1);
    chunk_info_t ci = get_chunk_info(disp_context);
    init_chunk_elems(&ci);

    for (int j = 0; j < N_ITERATIONS; ++j) {
        compute_next_chunk_elems(disp_context, fut_context, &ci);
        int8_t *sca = get_data(disp_context);
    }
    /* ... Free resources ... */
}
```

Finally, a C file `main.c` is needed to create the program's entry point. We initialize the MPI and Futhark environment. Then, our library, via the function `dispatch_context_new`, by specifying the size of the cellular automaton (600), its data type by using a predefined type in MPI (`MPI_INT8_T`), and the number of dimensions (one in this case). It returns a dispatch context that will need to be provided when calling other functions of our library.

The function `get_chunk_info` is called to get the chunk of the cellular automaton attributed to the current rank. Thus, we initiate it with the values that we want with the function `init_chunk_elems`.

In the temporal loop, we update our chunk using the created function `compute_next_chunk_elems`. In this function, we call the (+API) function `get_chunk_with_envelope` from our library with the following parameters:

* the dispatch context, obtained from `dispatch_context_new`,
* the Futhark context, obtained from `futhark_context_new`,
* the Chebyshev distance,
* and a pointer to a Futhark function that converts a C array to a Futhark array.

Our (+API) function handles the (+MPI) communication to exchange each chunk's missing and needed neighbors to create the envelope of the current process's chunk.

Thus, we call our Futhark function `futhark_entry_next_chunk_elems` to compute the new values. Given that the function is asynchronous, the function `futhark_context_sync` waits for the action to finish. Finally, we retrieve the new values with `futhark_values_i8_1d`.

After all that, it is possible to retrieve the entire cellular automaton on the root node by calling the function `get_data` on each task. The variable `sca` is not `NULL` if it is the root node, and it points to the entire cellular automaton.

## CPU Benchmarks

We perform benchmarks to validate the scalability of our one-dimensional distribution when compiling in sequential, multicore, (+OpenCL), or (+CUDA) mode. The benchmarks are performed on the (+HES-GE) cluster (Baobab/Yggdrasil).
The sequential and multicore benchmarks are performed as follows:

* the cellular automaton is $300,000,000$ cells in size,
* the number of tasks varies between $2^0$ and $2^7$,
* 15 measurements are performed, one measurement corresponds to one iteration,
* the iteration is computed 100 times.

| Number of tasks | Average [s] | Standard Derivation [s] | Speedup | Number of measures |
|:---:|:---:|:---:|:---:|:---:|
| 1 | 657.866 [s] | ± 14.977 [s] | x1.0 | 15 |
| 2 | 332.771 [s] | ± 2.814 [s] | x2.0 | 15 |
| 4 | 161.963 [s] | ± 7.309 [s] | x4.1 | 15 |
| 8 | 87.602 [s] | ± 2.918 [s] | x7.5 | 15 |
| 16 | 42.743 [s] | ± 0.039 [s] | x15.4 | 15 |
| 32 | 20.938 [s] | ± 0.007 [s] | x31.4 | 15 |
| 64 | 11.071 [s] | ± 0.024 [s] | x59.4 | 15 |
| 128 | 5.316 [s] | ± 0.191 [s] | x123.7 | 15 |
Table: Results for the distributed-sequential version of (+SCA)

This table contains the results obtained by using the backend `c` of Futhark.

| Number of tasks | Average [s] | Standard Derivation [s] | Speedup | Number of measures |
|:---:|:---:|:---:|:---:|:---:|
| 1 | 708.689 [s] | ± 16.036 [s] | x1.0 | 15 |
| 2 | 358.007 [s] | ± 4.037 [s] | x2.0 | 15 |
| 4 | 138.523 [s] | ± 3.773 [s] | x5.1 | 15 |
| 8 | 71.077 [s] | ± 1.280 [s] | x10.0 | 15 |
| 16 | 34.697 [s] | ± 0.834 [s] | x20.4 | 15 |
| 32 | 25.776 [s] | ± 0.725 [s] | x27.5 | 15 |
| 64 | 12.506 [s] | ± 0.554 [s] | x56.7 | 15 |
| 128 | 5.816 [s] | ± 0.045 [s] | x121.8 | 15 |
Table: Results for the distributed-multicore version of (+SCA)

This table contains the results obtained by using the backend `multicore` of Futhark.

\pagebreak

\cimgl{figs/elem_result_and_speedup_cpu.png}{width=\linewidth}{Benchmarks of the SCA in distributed-sequential/multicore}{Source: Realized by Baptiste Coudray}{fig:bench-cpu-sca}

We compare the average computation time for each number of tasks and each version (sequential and multicore) on the left graph. On the right graph, we compare the ideal speedup with the distributed-sequential and multicore version speedup.
The more we increase the number of tasks, the more the execution time is reduced. Thus, the distributed-sequential or multicore version speedup follows the curve of the ideal speedup. We can see that concurrent computing does not provide a significant performance gain over sequential computing because of the overhead of creating threads.

## GPU Benchmarks

The (+OpenCL) and (+CUDA) benchmarks are performed as follows:

* the cellular automaton has $300'000'000$ cells,
* the number of tasks varies between $2^0$ and $2^3$.
* 15 measurements are performed, one measurement corresponds to one iteration,
* the iteration is computed $50'000$ times,
* an NVIDIA GeForce RTX 3090 is allocated for each task.

| Number of tasks | Number of GPUs | Average [s] | Standard Derivation [s] | Speedup | Number of measures |
|:---:|:---:|:---:|:---:|:---:|:---:|
| 1 | 1 | 166.086 [s] | ± 0.096 [s] | x1.0 | 15 |
| 2 | 2 | 83.339 [s] | ± 0.099 [s] | x2.0 | 15 |
| 4 | 4 | 42.122 [s] | ± 0.078 [s] | x3.9 | 15 |
| 8 | 8 | 21.447 [s] | ± 0.031 [s] | x7.7 | 15 |
Table: Results for the distributed-(+OpenCL) version of (+SCA)

This table contains the results obtained by using the backend `opencl` of Futhark.

| Number of tasks | Number of GPUs | Average [s] | Standard Derivation [s] | Speedup | Number of measures |
|:---:|:---:|:---:|:---:|:---:|:---:|
| 1 | 1 | 160.291 [s] | ± 0.062 [s] | x1.0 | 15 |
| 2 | 2 | 80.434 [s] | ± 0.094 [s] | x2.0 | 15 |
| 4 | 4 | 40.640 [s] | ± 0.073 [s] | x3.9 | 15 |
| 8 | 8 | 20.657 [s] | ± 0.046 [s] | x7.8 | 15 |
Table: Results for the distributed-(+CUDA) version of (+SCA)

This table contains the results obtained by using the backend `cuda` of Futhark.

\cimgl{figs/elem_result_and_speedup_gpu.png}{width=\linewidth}{Benchmarks of the SCA in distributed-OpenCL/CUDA}{Source: Realized by Baptiste Coudray}{fig:bench-gpu-sca}

With this performance test (\ref{fig:bench-gpu-sca}), we compare the average computation time for each number of tasks/(+^GPU) and each version ((+OpenCL) and (+CUDA)) on the left graph. On the right graph, we compare the ideal speedup with the `distributed-opencl` and cuda version speedup. We notice that the computation time is essentially the same in (+OpenCL) as in (+CUDA). Moreover, the distributed follows the ideal speedup curve. Finally, we notice that parallel computation is up to four times faster than sequential/concurrent computation when executing with a single task/graphical card.

\pagebreak
