# Game of Life

The Game of Life is a zero-player game designed by John Horton Conway in 1970. It is also one of the best-known cellular automata. The game does not require the interaction of a player for it to evolve, it evolves thanks to these extremely simple rules:

1. a cell has eight neighbors,
2. a cell can be either alive or dead,
3. a dead cell with exactly three living neighbors becomes alive,
4. a living cell with two or three living neighbors stays alive; otherwise, it dies [@noauthor_jeu_2021].

## Example with the blinker

\cimg{figs/gol_blinker1.png}{scale=0.40}{First state of blinker}{Source: Taken from \url{https://commons.wikimedia.org/}, ref. URL04. Re-created by Baptiste Coudray}

A basic example is a blinker:

* the cell (one, one) and (one, three) die because they have seven dead neighbors and one living neighbor (rule n°4),
* the cell (zero, two) and (two, two) are born because they have three living neighbors (rule n°3),
* the cell (one, two) stays alive because it has two living neighbors (rule n°4).

\cimg{figs/gol_blinker2.png}{scale=0.40}{Second state of blinker}{Source: Taken from \url{https://commons.wikimedia.org/}, ref. URL05. Re-created by Baptiste Coudray}

Thus, after the application of the rules, the horizontal line becomes a vertical line. Then, at the next iteration, the vertical line becomes a horizontal line again.

## Distributed version

We create the game of life with our library to test it with a two-dimensional cellular automaton.
The code is relatively the same as the previous example; therefore, it is not explained, but you can find it in the Git repository.

```futhark
let count_neighbours [n][m] (board: [n][m]i8) :[n][m]i8 = ...
let compute_next_board [n][m] (chunk_board :[n][m]i8) 
                              (neighbours: [n][m]i8) :[][]i8 = ...

entry next_chunk_board [n][m] (chunk_board :[n][m]i8) :[][]i8 =
    let neighbours = count_neighbours chunk_board
    let next_board = compute_next_board chunk_board neighbours
    in next_board[1:n-1, 1:m-1]
```
This is a sneak peek of the `gol.fut` file. Like the (+SCA), we only update our cellular automaton and return it without the envelope.

```c
void compute_next_chunk_board(struct dispatch_context *dc, 
        struct futhark_context *fc, chunk_info_t *ci) { 
    struct futhark_i8_2d *fut_chunk_with_envelope = 
            get_chunk_with_envelope(dc, fc, 1, futhark_new_i8_2d);
    struct futhark_i8_2d *fut_next_chunk_board;
    futhark_entry_next_chunk_board(fc, &fut_next_chunk_board,
                                   fut_chunk_with_envelope);
    /* ... Sync, Get values & Free resources ... */
}

int main(int argc, char *argv[]) {
    /* ... MPI & Futhark Init ... */
    int board_dimensions[2] = {800, 800};
    struct dispatch_context *disp_context = 
            dispatch_context_new(board_dimensions, MPI_INT8_T, 2);
    /* ... */
    const int N_ITERATIONS = 100;
    for (int i = 0; i < N_MEASURES; ++i) { 
        compute_next_chunk_board(disp_context, fut_context, &ci);
    }
    /* ... Free resources ... */
}
```

In the C file `main.c`, we can see this almost the same code as for the (+SCA) example, but when calling `dispatch_context_new` we specify that this is a two-dimensional cellular automaton. In the `compute_next_chunk_board` function, we call `get_chunk_with_envelope` from our library with a different conversion function. It transforms a C 2D array to a Futhark 2D array.

## CPU Benchmarks

We perform benchmarks to validate the scalability of our two-dimensional distribution when compiling in sequential, multicore, (+OpenCL), or (+CUDA) mode. The benchmarks are performed on the (+HES-GE) cluster (Baobab/Yggdrasil).

The sequential and multicore benchmarks are performed as follows:

* the cellular automaton is $900'000'000$ cells in size,
* the number of tasks varies between $2^0$ and $2^7$,
* 15 measurements are performed, one measurement corresponds to one iteration,
* the iteration is computed 100 times.

| Number of tasks | Average [s] | Standard Derivation [s] | Speedup | Number of measures |
|:---:|:---:|:---:|:---:|:---:|
| 1 | 3471.723 [s] | ± 47.092 [s] | x1.0 | 15 |
| 2 | 1140.064 [s] | ± 56.780 [s] | x3.0 | 15 |
| 4 | 790.365 [s] | ± 10.501 [s] | x4.4 | 15 |
| 8 | 398.093 [s] | ± 13.438 [s] | x8.7 | 15 |
| 16 | 221.687 [s] | ± 4.152 [s] | x15.7 | 15 |
| 32 | 100.422 [s] | ± 0.068 [s] | x34.6 | 15 |
| 64 | 55.986 [s] | ± 1.587 [s] | x62.0 | 15 |
| 128 | 28.111 [s] | ± 0.263 [s] | x123.5 | 15 |
Table: Results for the distributed-sequential version of Game of Life

This table contains the results obtained by using the backend `c` of Futhark.

| Number of tasks | Average [s] | Standard Derivation [s] | Speedup | Number of measures |
|:---:|:---:|:---:|:---:|:---:|
| 1 | 2154.686 [s] | ± 198.122 [s] | x1.0 | 15 |
| 2 | 1160.921 [s] | ± 77.230 [s] | x1.9 | 15 |
| 4 | 502.860 [s] | ± 3.465 [s] | x4.3 | 15 |
| 8 | 206.818 [s] | ± 4.179 [s] | x10.4 | 15 |
| 16 | 106.103 [s] | ± 0.450 [s] | x20.3 | 15 |
| 32 | 71.463 [s] | ± 0.485 [s] | x30.2 | 15 |
| 64 | 39.116 [s] | ± 0.489 [s] | x55.1 | 15 |
| 128 | 14.008 [s] | ± 0.335 [s] | x153.8 | 15 |
Table: Results for the distributed-multicore version of Game of Life

This table contains the results obtained by using the backend `multicore` of Futhark.

\cimgl{figs/gol_result_and_speedup_cpu.png}{width=\linewidth}{Benchmarks of the game of life in distributed-sequential/multicore}{Source: Realized by Baptiste Coudray}{fig:bench-cpu-gol}

We notice an apparent difference between the distributed-sequential and multicore version when there is only one task. The multicore version is $1.6$ times faster than the sequential version. Nevertheless, both versions have a perfect speedup. The multicore version even gets a maximum speedup of x154 with 128 tasks. This performance can be explained by the caching of data in the processor and the use of threads.

## GPU Benchmarks

The (+OpenCL) and (+CUDA) benchmarks are performed as follows:

* the cellular automaton has $900'000'000$ cells,
* the number of tasks varies between $2^0$ and $2^3$.
* 15 measurements are performed, one measurement corresponds to one iteration,
* the iteration is computed $8'000$ times,
* an NVIDIA GeForce RTX 3090 is allocated for each task.

| Number of tasks | Number of GPUs | Average [s] | Standard Derivation [s] | Speedup | Number of measures |
|:---:|:---:|:---:|:---:|:---:|:---:|
| 1 | 1 | 230.144 [s] | ± 0.225 [s] | x1.0 | 15 |
| 2 | 2 | 115.400 [s] | ± 0.070 [s] | x2.0 | 15 |
| 4 | 4 | 58.019 [s] | ± 0.104 [s] | x4.0 | 15 |
| 8 | 8 | 29.157 [s] | ± 0.061 [s] | x7.9 | 15 |
Table: Results for the distributed-(+OpenCL) version of Game of Life

This table contains the results obtained by using the backend `opencl` of Futhark.

| Number of tasks | Number of GPUs | Average [s] | Standard Derivation [s] | Speedup | Number of measures |
|:---:|:---:|:---:|:---:|:---:|:---:|
| 1 | 1 | 218.807 [s] | ± 0.057 [s] | x1.0 | 15 |
| 2 | 2 | 109.598 [s] | ± 0.109 [s] | x2.0 | 15 |
| 4 | 4 | 55.039 [s] | ± 0.100 [s] | x4.0 | 15 |
| 8 | 8 | 27.737 [s] | ± 0.050 [s] | x7.9 | 15 |
Table: Results for the distributed-(+CUDA) version of Game of Life

This table contains the results obtained by using the backend `cuda` of Futhark.

\pagebreak

\cimgl{figs/gol_result_and_speedup_gpu.png}{width=\linewidth}{Benchmarks of the game of life in distributed-OpenCL/CUDA}{Source: Realized by Baptiste Coudray}{fig:bench-gpu-gol}

With this performance test (\ref{fig:bench-gpu-gol}), we notice that the computation time is essentially the same in (+OpenCL) as in (+CUDA). Moreover, the distribution follows the ideal speedup curve. Furthermore, we notice that parallel computation is up to 15 times faster than sequential/concurrent computation when executing with a single task/graphical card.

\pagebreak
