# Lattice-Boltzmann Method

"_The lattice Boltzmann method (LBM) has established itself in the past decades as a valuable approach to Computational
Fluid Dynamics (CFD). It is commonly used to model time-dependent, incompressible or compressible flows in a
regime of Direct Numerical Simulation (DNS) or Large Eddy Simulation (LES). One of its strengths lies in the ability to easily
represent complex physical phenomena, ranging from multi-phase flows to reactive and suspension flows. The
method originates in a molecular description of a fluid, based on the Boltzmann equation, and can directly incorporate
physical terms stemming from a knowledge of the interaction between molecules. It is therefore an invaluable tool in
fundamental research, as it keeps the cycle between the elaboration of a theory and the formulation of a corresponding numerical model short._" [@latt_palabos_2020]

## Distributed version

We implement the Lattice-Boltzmann Method with our library to test it with a three-dimensional cellular automaton. The code is almost the same as the previous example, so we focus on custom types in this example. A cell in this cellular automaton is of type of array containing 27 floats. In MPI, there is no predeclared type of array, so we need to create one. Thanks to our library, it is possible to do that with a minimum code.

```c
typedef struct lbm_values {
    float values[NB_VALUES];
} lbm_values_t;

int main(int argc, char *argv[]) {
    /* ... MPI & Futhark Init ... */
    int count = 1;
    int block_lengths[] = {NB_VALUES};
    MPI_Aint displacements[] = {offsetof(struct lbm_values, values)};
    MPI_Datatype types[] = {MPI_FLOAT};
    
    MPI_Datatype lbm_type = create_type(count, block_lengths, displacements, 
                                        types);

    int lbm_dimensions[3] = {400, 400, 400};
    struct dispatch_context *disp_context = 
            dispatch_context_new(lbm_dimensions, lbm_type, 3);
    /* ... Temporal loop & Free resources ... */
}
````

We need to declare a structure specifying a cell, so `lbm_values` is a struct containing an array of 27 floats. After that, in the `main` function, we describe the structure as follows:

* there is one field (`values`),
* the first field is containing 27 values,
* we compute where is located the field `values` in the structure,
* `values` is of type `MPI_FLOAT`.

Finally, we call our library function `create_type` with the previous parameters, and it returns an `MPI_Datatype` that can be pass to the function `dispatch_context_new`.

## CPU Benchmarks

We perform benchmarks to validate the scalability of our three-dimensional distribution when compiling in sequential, multicore, (+OpenCL), or (+CUDA) mode. The benchmarks are performed on the (+HES-GE) cluster (Baobab/Yggdrasil).
The sequential and multicore benchmarks are performed as follows:

* the cellular automaton is $27'000'000$ cells in size,
* the number of tasks varies between $2^0$ and $2^7$,
* 15 measurements are performed, one measurement corresponds to one iteration,
* the iteration is computed 100 times.

| Number of tasks | Average [s] | Standard Derivation [s] | Speedup | Number of measures |
|:---:|:---:|:---:|:---:|:---:|
| 1 | 716.133 [s] | ± 5.309 [s] | x1.0 | 15 |
| 2 | 363.166 [s] | ± 3.482 [s] | x2.0 | 15 |
| 4 | 185.430 [s] | ± 0.847 [s] | x3.9 | 15 |
| 8 | 93.994 [s] | ± 0.566 [s] | x7.6 | 15 |
| 16 | 81.266 [s] | ± 8.947 [s] | x8.8 | 15 |
| 32 | 41.040 [s] | ± 1.590 [s] | x17.4 | 15 |
| 64 | 22.188 [s] | ± 0.321 [s] | x32.3 | 15 |
| 128 | 17.415 [s] | ± 4.956 [s] | x41.1 | 15 |
Table: Results for the distributed-sequential version of (+LBM)

This table contains the results obtained by using the backend `c` of Futhark.

| Number of tasks | Average [s] | Standard Derivation [s] | Speedup | Number of measures |
|:---:|:---:|:---:|:---:|:---:|
| 1 | 695.675 [s] | ± 8.867 [s] | x1.0 | 15 |
| 2 | 352.925 [s] | ± 4.293 [s] | x2.0 | 15 |
| 4 | 181.736 [s] | ± 0.695 [s] | x3.8 | 15 |
| 8 | 237.983 [s] | ± 0.271 [s] | x2.9 | 15 |
| 16 | 79.360 [s] | ± 2.185 [s] | x8.8 | 15 |
| 32 | 46.285 [s] | ± 0.138 [s] | x15.0 | 15 |
| 64 | 24.059 [s] | ± 0.061 [s] | x28.9 | 15 |
| 128 | 16.614 [s] | ± 1.088 [s] | x41.9 | 15 |
Table: Results for the distributed-multicore version of (+LBM)

This table contains the results obtained by using the backend `multicore` of Futhark.

\pagebreak

\cimgl{figs/lbm_result_and_speedup_cpu.png}{width=\linewidth}{Benchmarks of the LBM in distributed-sequential/multicore}{Source: Realized by Baptiste Coudray}{fig:bench-cpu-lbm}

Contrary to the previous benchmarks, the speedups do not follow the ideal speedup curve. Indeed, whether in sequential or multicore, we obtain a maximum speedup with 128 tasks of x41 when we were hoping to have a speedup of x128.

## GPU Benchmarks

The (+OpenCL) and (+CUDA) benchmarks are performed as follows:

* the cellular automaton has $27'000'000$ cells,
* the number of tasks varies between $2^0$ and $2^3$.
* 15 measurements are performed, one measurement corresponds to one iteration,
* the iteration is computed $3'000$ times,
* an NVIDIA GeForce RTX 3090 is allocated for each task.

| Number of tasks | Number of GPUs | Average [s] | Standard Derivation [s] | Speedup | Number of measures |
|:---:|:---:|:---:|:---:|:---:|:---:|
| 1 | 1 | 210.347 [s] | ± 0.096 [s] | x1.0 | 15 |
| 2 | 2 | 99.677 [s] | ± 0.038 [s] | x2.1 | 15 |
| 4 | 4 | 40.710 [s] | ± 0.076 [s] | x5.2 | 15 |
| 8 | 8 | 20.800 [s] | ± 0.031 [s] | x10.1 | 15 |
Table: Results for the distributed-(+OpenCL) version of (+LBM)

This table contains the results obtained by using the backend `opencl` of Futhark.

| Number of tasks | Number of GPUs | Average [s] | Standard Derivation [s] | Speedup | Number of measures |
|:---:|:---:|:---:|:---:|:---:|:---:|
| 1 | 1 | 207.683 [s] | ± 0.249 [s] | x1.0 | 15 |
| 2 | 2 | 99.177 [s] | ± 0.056 [s] | x2.1 | 15 |
| 4 | 4 | 40.240 [s] | ± 0.074 [s] | x5.2 | 15 |
| 8 | 8 | 20.459 [s] | ± 0.037 [s] | x10.2 | 15 |
Table: Results for the distributed-(+CUDA) version of (+LBM)

This table contains the results obtained by using the backend `cuda` of Futhark.

\cimgl{figs/lbm_result_and_speedup_gpu.png}{width=\linewidth}{Benchmarks of the LBM in distributed-OpenCL/CUDA}{Source: Realized by Baptiste Coudray}{fig:bench-gpu-lbm}

Like the other benchmarks (\ref{fig:bench-gpu-sca}, \ref{fig:bench-gpu-gol}), there is very little difference between the (+OpenCL) and (+CUDA) versions (computation time and speedup). We get a more than ideal speedup with 2, 4, and 8 tasks/(+^GPU) (x2.1, x5.2, and x10.2, respectively). Finally, we notice that parallel computation is up to 3 times faster than sequential/concurrent computation when executing with a single task/graphical card.

\pagebreak
